
#include "../blitz.h"

#ifdef BB_GC_RC

//#define DEBUG_GC

#define ALIGN_SIZE				BBMEM_ALIGNSIZE
#define ALIGN_MASK				BBMEM_ALIGNMASK

#define BUF_GROW				1024

#define BBGC_ZERO				0x80000000

#define PAGE_BITS				12
#define PAGE_SIZE				(1 << PAGE_BITS)
#define PAGE_MASK				((PAGE_SIZE / 4) - 1)

static BBGCMode GcMode			= BBGC_AUTOMATIC;
static BBGCDebugMode GcDebugMode = BBGC_NODEBUG;

static BBSize GcSuspended		= 0;
static int gc_millisecs			= 0;
static int gc_memfreed			= 0;
static int gc_objsfreed			= 0;
static int gc_alloced			= 0;

static void **BufBase;
static void **BufPut;
static void **BufEnd;

static size_t MaxBufLen;

static int* Pages[1 << (25 - PAGE_BITS)];
static unsigned int PagesAlloced;

static void **Live;
static size_t LiveLen;

#define MAXSIZE					256
#define MAXFREEBUFSIZE			65536

static char* FreeBuf;
static size_t FreeBufSize;
static int FreeBufAlloced;
static char* FreeLists[MAXSIZE / ALIGN_SIZE];

#ifdef DEBUG_GC

#	define DEBUG_COUNT				16384	// 4096
#	define DEBUG_MASK				(DEBUG_COUNT - 1)
#	define DEBUG_OBJECT(obj, msg)	debugObject((obj), (msg))

typedef struct GCDebugInfo		GCDebugInfo;

struct GCDebugInfo
{
	void			*object;
	void			*typename;
	const char		*msg;
};

static size_t DebugIndex;
static GCDebugInfo DebugInfos[DEBUG_COUNT];

static void dumpDebugInfo();

static bool DebugAtExit;

static const char* object_get_class_name(const BBObject *obj)
{
	BBDebugScope *scope = obj->clas->debug_scope;
	return scope ? scope->name : "?";
}

static void debugObject(void *ptr, const char *msg)
{
	GCDebugInfo *db = DebugInfos + (DebugIndex & DEBUG_MASK);
	db->object		= ptr;
	db->typename	= object_get_class_name((BBObject *)ptr);
	db->msg			= msg;
	
	++DebugIndex;
	if (DebugAtExit) {
		return;
	}
	
	atexit(dumpDebugInfo);
	DebugAtExit = true;
}

static void dumpDebugInfo()
{
	printf("GC Debug info:\n");

	for (size_t i = 0; i < DEBUG_COUNT; ++i)
	{
		GCDebugInfo *db = DebugInfos + ((DebugIndex + i) & DEBUG_MASK);
		if (!db->object) {
			continue;
		}
		printf("%s %s @ $%p\n", db->msg, db->typename, db->object);
	}
	
	fflush(stdout);
}

#else
#	define DEBUG_OBJECT(obj, msg)
#endif

static void gc_error(void *ptr, const char *msg)
{
#ifdef DEBUG_GC
	printf("GC ERROR: %s, object=$%p\n", msg, ptr);
	fflush(stdout);	
#endif
	bbExThrowCString(msg);
}

//	previously named setMemBit
static int gc_set_membit(void *ptr)
{
	int page = (unsigned)ptr >> (PAGE_BITS + 7);

	if (!Pages[page]) {
		++PagesAlloced;
		Pages[page] = malloc(PAGE_SIZE);
		memset(Pages[page], 0, PAGE_SIZE);
	}

	int offset = ((unsigned)ptr >> 9) & PAGE_MASK;
	int mask = 1 << (((unsigned)ptr >> 4) & 31);
	
	if (Pages[page][offset] & mask) {
		gc_error(ptr, "setMemBit error: membit already set");
	}
		
	Pages[page][offset] |= mask;
	
	DEBUG_OBJECT(ptr, "setMemBit");
}

//	previously named clrMemBit
static void gc_clear_membit(void *ptr)
{
	int page = (unsigned)ptr >> (PAGE_BITS+7);

	if (!Pages[page]) {
		gc_error(ptr, "clrMemBit error: mempage does not exist");
	}
		
	int offset = ((unsigned)ptr >> 9) & PAGE_MASK;
	int mask = 1 << (((unsigned)ptr >> 4) & 31 );
	
	if (!(Pages[page][offset] & mask)) {
		gc_error(ptr, "clrMemBit error: membit not set");
	}
		
	Pages[page][offset] &= ~mask;
	
	DEBUG_OBJECT(ptr, "clrMemBit");
}

//	previously named tstMemBit
static int gc_test_membit(const void *ptr)
{
	if ((int)ptr & 15) {
		return 0;
	}
		
	int page = (unsigned)ptr >> (PAGE_BITS + 7);
	if (!Pages[page]) {
		return 0;
	}
	
	int offset = ((unsigned)ptr >> 9) & PAGE_MASK;
	int mask = 1 << (((unsigned)ptr >> 4) & 31);
	
	return Pages[page][offset] & mask;
}

//	previously named collectMem
static int gc_collect_mem()
{
	void* root_regs[BBGC_NUM_ROOTREGS];
	
	if (GcSuspended || !bbGCStackTop) {
		return 0;
	}
		
	++GcSuspended;
	
	gc_memfreed = gc_alloced;
	
	if (GcDebugMode) {
		gc_millisecs = -1;
#if BB_OS_WIN32
		gc_millisecs = timeGetTime();
#endif
	}

	void **spBase = bbGCRootRegs(root_regs);
	
	int n = bbGCStackTop - spBase + BBGC_NUM_ROOTREGS;
	if (n > LiveLen)
	{
		void **tmp = Live;
		int new_len = LiveLen + 1000;
		if (n > new_len) {
			new_len = n;
		}
		
		Live = (void **)malloc(new_len * 4);
		
		if (tmp) {
			free(tmp);
		}
		
		LiveLen = new_len;
	}

	BBObject *obj;
	
	void **tsp = Live;
	void **sp = spBase;
	while (sp != bbGCStackTop)
	{
		obj = *sp++;
		if (gc_test_membit(obj)) {
			*tsp++ = obj;
		}
	}
	
	sp = root_regs;
	
	for (n = 0; n < BBGC_NUM_ROOTREGS; ++n)
	{
		obj = *sp++;
		if (gc_test_membit(obj)) {
			*tsp++ = obj;
		}
	}
	
	sp = Live;
	while (sp != tsp) {
		obj = *sp++;
		BB_INCREFS(obj);
	}
	
	gc_objsfreed = 0;

	//int i = 0;
	//while (i != (BufPut - BufBase))
	for (int obj_id = 0; obj_id != (BufPut - BufBase); ++obj_id)
	{
		//obj = BufBase[i++];
		obj = BufBase[obj_id];
		
		// Check for incorrectly freed objects.
		if (obj->refs >= 0) {
			
#		ifdef	BB_DEBUG
			printf("Error: Bad references | ");
			printf("obj = $%x (%s) refs = $%x\n",
				   (unsigned)obj, obj->clas->debug_scope->name, (unsigned)obj->refs);

			if (obj->clas == &bbStringClass) {
				BBString *str = (BBString *)str;
				
				printf(" .length = %i\n", str->length);
				
				char *buf = bbStringToCString(str);
				printf(" .buf = \q%w\q\n", buf);
				bbMemFree(buf);
			}
			
			if (obj->clas == &bbArrayClass) {
				BBArray *arr = (BBArray *)obj;
				
				printf(" .type = %s\n", arr->type);
				printf(" .dims = %i\n", arr->dims);
				printf(" .size = %i\n", arr->size);
				printf(" .scales[] = { %i", arr->scales[0]);
								
				for (BBSize dim = 1; dim < arr->dims; ++dim) {
					printf(", %i", arr->scales[dim]);
				}
				
				printf(" }\n");
			}

			fflush(stdout);
			
			bbOnDebugStop();
			
#		else
			printf("bad refs:obj=$%x refs=$%x\n", (unsigned)obj, (unsigned)obj->refs);
			fflush(stdout);
#		endif
		}
		
		if (obj->refs &= ~BBGC_ZERO) {
			continue;
		}
		
		obj->clas->free(obj);
		++gc_objsfreed;
	}
	
	BufPut = BufBase;

	sp = Live;
	while (sp != tsp)
	{
		obj = *sp++;
		BB_DECREFS(obj);
	}
	
	gc_memfreed -= gc_alloced;
	
	if (GcDebugMode) {
#if BB_OS_WIN32
		gc_millisecs = timeGetTime() - gc_millisecs;
#endif
		printf(
			"GC collectMem: memFreed=%i, time=%ims, objsFreed=%i, objsScanned=%i, objsLive=%i\n",
			gc_memfreed,
			gc_millisecs,
			gc_objsfreed,
			bbGCStackTop - spBase,
			tsp - Live
		);
		fflush(stdout);
	}
	
	--GcSuspended;
	
	return gc_memfreed;
}

// API //

void bbGCStartup() {}

BBGCMode bbGCGetMode()
{
	return GcMode;
}

void bbGCSetMode(BBGCMode mode)
{
	if (mode == BBGC_AUTOMATIC || mode == BBGC_MANUAL || mode == BBGC_AGGRESSIVE) {
		GcMode = mode;
	} else {
		GcMode = BBGC_AUTOMATIC;
	}
}

void bbGCSetDebug(BBGCDebugMode mode)
{
	if (mode == BBGC_NODEBUG || mode == BBGC_STDOUTDEBUG) {
		GcDebugMode = mode;
	} else {
		GcDebugMode = BBGC_NODEBUG;
	}
}

//	param: flags is unused atm
BBObject* bbGCAllocObject(const BBSize size, const BBClass *clas, int flags)
{
	BB_ASSERT(size >= sizeof (BBObject), "The object's size is too small.");
	BB_ASSERT(size <= BBSIZE_MAX, "The object's size is too large.");
	
	BBObject *obj;

	if (GcMode == BBGC_AUTOMATIC && !GcSuspended)
	{
		static size_t alloced;
		static int rate = 500;
		
		alloced += size;
		
		if (alloced > (1024 * 1024) || BufPut - BufBase > rate) {
			gc_collect_mem();
			rate += 500 - gc_objsfreed;
			alloced = 0;
		}
	} else if (GcMode == BBGC_AGGRESSIVE) {
		gc_collect_mem();
	}

	if (size < 1) {
		return NULL;
	}
		
	if (size > (MAXSIZE - ALIGN_SIZE)) {
		obj = bbMemAlloc(size);
	} else
	{
		int n = (size + ALIGN_MASK) / ALIGN_SIZE;
		obj = FreeLists[n];
	
		if (obj) {
			FreeLists[n] = *(char **)obj;
		} else if (size <= FreeBufSize) {
			obj = FreeBuf;
			n *= ALIGN_SIZE;
			FreeBuf		+= n;
			FreeBufSize	-= n;
		} else {
			if (FreeBufSize) {
				int n = (FreeBufSize + ALIGN_MASK) / ALIGN_SIZE;
				*(char **)FreeBuf = FreeLists[n];
				FreeLists[n] = FreeBuf;
			}
			
			obj = bbMemAlloc(MAXFREEBUFSIZE);
			n *= ALIGN_SIZE;
			FreeBuf = (char *)obj + n;
			FreeBufSize = MAXFREEBUFSIZE - n;
		}
	}
	
	gc_set_membit(obj);
	gc_alloced += size;
	
	obj->clas = clas;
	obj->refs = 0;
	
	bbGCFree(obj);
	return obj;
}

BBBool bbGCValidate(const void *ptr)
{
	return gc_test_membit(ptr) ? BBTRUE : BBFALSE;
}

BBInt bbGCMemAlloced()
{
	return gc_alloced;
}

BBInt bbGCCollect()
{
	return gc_collect_mem();
}

void bbGCSuspend()
{
	if (GcSuspended < BBGC_MAX_SUSPENDED) {
		++GcSuspended;
	}
}

BBBool bbGCIsSuspended()
{
	return !GcSuspended;
}

void bbGCResume()
{
	if (GcSuspended) {
		--GcSuspended;
	}
}

//	same as BB_INCREFS(obj)
void bbGCRetain(BBObject *obj)
{
	++obj->refs;
}

//	same as BB_DECREFS(obj)
void bbGCRelease(BBObject *obj)
{
	if (!--obj->refs) {
		bbGCFree(obj);
	}
}

// Reference counting GC only
void bbGCFree(BBObject *obj)
{
	if (obj->refs) {
		bbExThrowCString("GC bbGCFree: mem has non-0 refs");
	}
	
	if (BufPut == BufEnd)
	{
		int len	= BufPut - BufBase;
		int new_len	= len + BUF_GROW;
		
		BufBase = (void **)bbMemExtend(BufBase, len * 4, new_len * 4);
		BufEnd = BufBase + new_len;
		BufPut = BufBase + len;
		
		if (new_len > MaxBufLen) {
			MaxBufLen = new_len;
		}
	}
	
	obj->refs = BBGC_ZERO;
	*BufPut++ = obj;
}

void bbGCDeallocObject(BBObject *obj, BBSize size)
{
	if (!obj) {
		return;
	}
	
	gc_clear_membit(obj);
	
	if (size > (MAXSIZE - ALIGN_SIZE)) {
		bbMemFree(obj);
	} else {
		int n = (size + ALIGN_MASK) / ALIGN_SIZE;
		*(char **)obj = FreeLists[n];
		FreeLists[n] = (char *)obj;
	}
	
	gc_alloced -= size;
}

#endif	// BB_GC_RC
