#include "blitz_string.h"
#include "blitz_unicode.h"
#include "blitz_array.h"

#include "exceptions/blitz_runtimeexception.h"

typedef uint_fast32_t	arrpos_t;
typedef uint_fast32_t	strpos_t;

static void string_free(BBString *str);

static BBDebugScope ClassScope =
{
	.kind			= BBDEBUGSCOPE_USERTYPE,
	.name			= "String",
	.decls			= {
		BBDEBUGDECL_END
	}
};

BBClass bbStringClass =
{
	.super			= &bbObjectClass,
	.free			= string_free,
	.debug_scope	= &ClassScope,
	.instance_size	= 0,
	
	.ctor			= 0,
	.dtor			= 0,
	
	.ToString		= bbStringToString,
	.Compare		= bbStringCompare,
	.SendMessage	= bbObjectSendMessage,
	.Clone			= bbStringClone,
	
	._reserved1_	= bbObjectReserved,
	._reserved2_	= bbObjectReserved,
	
	.vfns			=
	{	
		bbStringFind,
		bbStringFindLast,
		bbStringTrim,
		bbStringReplace,
		
		bbStringToLower,
		bbStringToUpper,
		
		bbStringToInt,
		bbStringToLong,
		bbStringToFloat,
		bbStringToDouble,
		bbStringToCString,
		bbStringToWString,
		
		bbStringFromInt,
		bbStringFromLong,
		bbStringFromFloat,
		bbStringFromDouble,
		bbStringFromCString,
		bbStringFromWString,
		
		bbStringFromBytes,
		bbStringFromShorts,
		
		bbStringStartsWith,
		bbStringEndsWith,
		bbStringContains,
		
		bbStringSplit,
		bbStringJoin,
		
		bbStringFromUTF8String,
		bbStringToUTF8String
	}
};

BBString bbEmptyString =
{
	.clas	= &bbStringClass,
	.refs	= BBGC_MANYREFS,
	.length	= 0
};

// Not called in threaded mode.
static void string_free(BBString *str)
{
#ifdef BB_GC_RC
	if (str == &bbEmptyString) {
		str->refs = BBGC_MANYREFS;
		return;
	}
	
	bbGCDeallocObject(str, sizeof (BBString) + str->length * sizeof (BBChar));
#endif
}


//	INTERNAL //

extern BBBool bb_strbuf_equals(const BBChar * restrict str_buf,
							   const BBChar * restrict cmp_buf,
							   const BBSize len);

BBSize bb_strbuf_findall(str_buf, str_len, sub_buf, sub_len, matches)
 const BBChar * restrict str_buf;
 const BBSize str_len;
 const BBChar * restrict sub_buf;
 const BBSize sub_len;
 BBSize * const matches;
{
	if (str_len < sub_len) {
		return 0;
	}
	
	--str_buf;
	
	BBSize num_matches		= 0;
	const BBChar sub_first	= sub_buf[0];
	
	for (strpos_t sub_pos, str_pos = 0; str_pos < str_len; ++str_pos, ++str_buf)
	{
		// Check if first characters are a match.
		if (*str_buf == sub_first) {
		
			for (sub_pos = 1; sub_pos < sub_len; ++sub_pos)
			{
				if (str_buf[sub_pos] != sub_buf[sub_pos]) {
					break;
				}
			}
			
			// Found a complete match.
			if (sub_pos == sub_len) {
				
				if (matches) {
					matches[num_matches] = str_pos;
				}
				
				++num_matches;
			}
		}
	}
	
	return num_matches;
}


// Compile time exception messages

#define CHAR_OUTOFUCS2RANGEBUF_LEN		36
static const BBChar CharOutOfUCS2RangeBuffer[CHAR_OUTOFUCS2RANGEBUF_LEN] = {
	'U','n','i','c','o','d','e',' ', 'c','h','a','r','a','c','t','e','r',' ',
	'o','u','t',' ','o','f',' ','U','C','S','-','2',' ', 'r','a','n','g','e'
};


//	API //

BBString* bbStringNew(BBInt len)
{
	BB_ASSERT(len < BBSTRING_MAX_LENGTH, "String is too long.");
	
	if (len < 1) {
		return &bbEmptyString;
	} else if (len > BBSTRING_MAX_LENGTH) {
		len = BBSTRING_MAX_LENGTH;
	}
	
	BBString *str	= BBSTRING_ALLOC(len);
	str->length		= len;
	
	return str;
}

BBString* bbStringFromChar(BBInt codepoint)
{	
	if (codepoint & 0xffff0000) {
		BBRuntimeException *ex;
		
		ex = bbRuntimeExceptionCreate(
				BBSTRING_FROMBUFFER(
					CharOutOfUCS2RangeBuffer,
					CHAR_OUTOFUCS2RANGEBUF_LEN
				),
				BBNULL
			 );
		
		bbExThrow(ex);
	}
	
	BBString *str	= BBSTRING_ALLOC(1);
	str->length		= 1;
	str->buf[0]		= codepoint;
	
	return str;
}

#ifdef BB_ENABLE_EXPERIMENTAL

#define INT_MINBUF_LEN			11
static const BBChar IntMinValueBuffer[INT_MINBUF_LEN] =
	{ '-','2','1','4','7','4','8','3','6','4','8' };

#define LONG_MINBUF_LEN			20
static const BBChar LongMinValueBuffer[LONG_MINBUF_LEN] =
	{ '-','9','2','2','3','3','7','2','0','3', '6','8','5','4','7','7','5','8','0','8' };

#endif	// BB_ENABLE_EXPERIMENTAL

BBString* bbStringFromInt(BBInt number)
{
	char buf[64];
	char *ptr = buf + 64;
	
	bool is_negative = number < 0;
	
	if (is_negative)
	{
		number = -number;
		
		if (number < 0) {
#		ifdef BB_ENABLE_EXPERIMENTAL
			return BBSTRING_FROMBUFFER(IntMinValueBuffer, INT_MINBUF_LEN);
#		else
			return bbStringFromBytes("-2147483648", 11);
#		endif
		}
	}
	
	do {
		*--ptr = number % 10 + '0';
	} while (number /= 10);
	
	if (is_negative) {
		*--ptr = '-';
	}
	
	return bbStringFromBytes(ptr, buf + 64 - ptr);
}

BBString* bbStringFromLong(BBLong number)
{
	char buf[64];
	char *ptr = buf + 64;
	
	bool is_negative = number < 0;
	
	if (is_negative)
	{
		number = -number;
		if (number < 0) {
#		ifdef BB_ENABLE_EXPERIMENTAL
			return BBSTRING_FROMBUFFER(LongMinValueBuffer, LONG_MINBUF_LEN);
#		else
			return bbStringFromBytes("-9223372036854775808", 20);
#		endif
		}
	}
	
	do {
		*--ptr = number % 10 + '0';
	} while (number /= 10);
	
	if (is_negative) {
		*--ptr = '-';
	}
	
	return bbStringFromBytes(ptr, buf + 64 - ptr);
}

BBString* bbStringFromFloat(BBFloat number)
{
	char buf[64];
	snprintf(buf, 64, "%#.9g", number);
	return bbStringFromCString(buf);
}

BBString* bbStringFromDouble(BBDouble number)
{
	char buf[64];
	snprintf(buf, 64, "%#.17lg", number);
	return bbStringFromCString(buf);
}

BBString* bbStringFromBytes(const char * const buf, BBInt size)
{
	BB_ASSERT(size < BBSTRING_MAX_LENGTH, "The resulting string would be too long.");
	
	if (size < 1) {
		return &bbEmptyString;
	} else if (size > BBSTRING_MAX_LENGTH) {
		size = BBSTRING_MAX_LENGTH;
	}
	
	BBString *str = BBSTRING_ALLOC(size);
	str->length = size;
	
	for (strpos_t i = 0; i < size; ++i) {
		str->buf[i] = (BBByte)buf[i];
	}
	
	return str;
}

BBString* bbStringFromShorts(const BBShort *buf, BBInt size)
{
	BB_ASSERT(size < BBSTRING_MAX_LENGTH, "The resulting string would be too long.");
	
	if (size < 1) {
		return &bbEmptyString;
	} else if (size > BBSTRING_MAX_LENGTH) {
		size = BBSTRING_MAX_LENGTH;
	}
	
	BBString *str = BBSTRING_ALLOC(size);
	str->length = size;
	
	memcpy(str->buf, buf, size * sizeof (BBShort));
	return str;
}

BBString* bbStringFromInts(const BBInt *buf, BBInt size)
{
	BB_ASSERT(size < BBSTRING_MAX_LENGTH, "The resulting string would be too long.");
	
	if (size < 1) {
		return &bbEmptyString;
	} else if (size > BBSTRING_MAX_LENGTH) {
		size = BBSTRING_MAX_LENGTH;
	}
	
	BBString *str = BBSTRING_ALLOC(size);
	str->length = size;
	
	for (strpos_t i = 0; i < size; ++i) {
		str->buf[i] = buf[i];
	}
	
	return str;
}

BBString* bbStringFromArray(const BBArray *arr)
{
	if (arr->dims != 1) {
		return &bbEmptyString;
	}
	
	const BBSize len = arr->scales[0];
	const void *data = BBARRAY_DATAPTR(arr);
	
	switch (arr->type[0]) {
	case 'b':
		return bbStringFromBytes(data, len);
	case 's':
		return bbStringFromShorts(data, len);
	case 'i':
		return bbStringFromInts(data, len);
	}
	
	return &bbEmptyString;
}

BBString* bbStringFromCString(const char *c_str)
{
	return c_str ? bbStringFromBytes(c_str, strlen(c_str)) : &bbEmptyString;
}

BBString* bbStringFromWString(const BBChar *w_str)
{
	if (!w_str) {
		return &bbEmptyString;
	}
	
	const BBChar *last_pos = w_str;
	
	while (*last_pos) {
		++last_pos;
	}
	
	return bbStringFromShorts(w_str, last_pos - w_str);
}

BBString* bbStringFromUTF8String(const char *utf8_str)
{
	if (!utf8_str) {
		return &bbEmptyString;
	}
	
	size_t utf8_len = strlen(utf8_str);
	
	if (!utf8_len) {
		return &bbEmptyString;
	} else if (utf8_len > BBSTRING_MAX_LENGTH) {
		utf8_len = BBSTRING_MAX_LENGTH;
	}
	
	BB_ASSERT(utf8_len < BBSTRING_MAX_LENGTH, "The resulting string would be too long.");
	
	BBChar *str_buf = malloc(utf8_len * sizeof (BBChar));
	BBChar *str_chr = str_buf;
	
	uint32_t utf8_code;
	while (utf8_code = *utf8_str++ & 0xff)
	{
		if (utf8_code < 0x80) {
			*str_chr++ = utf8_code;
		} else {
			int d = *utf8_str++ & 0x3f;
			
			if (utf8_code < 0xe0) {
				*str_chr++ = ((utf8_code & 31) << 6) | d;
			} else {
				int e = *utf8_str++ & 0x3f;
				
				if (utf8_code < 0xf0) {
					*str_chr++ = ((utf8_code & 15) << 12) | (d << 6) | e;
				} else {
					int f = *utf8_str++ & 0x3f;
					int v = ((utf8_code & 7) << 18) | (d << 12) | (e << 6) | f;
					
					if (v & 0xffff0000)
					{
						BBRuntimeException *ex;
#					ifdef BB_ENABLE_EXPERIMENTAL
						ex = bbRuntimeExceptionCreate(BBSTRING_FROMBUFFER(CharOutOfUCS2RangeBuffer, CHAR_OUTOFUCS2RANGEBUF_LEN), BBNULL);
#					else
						ex = bbRuntimeExceptionCreate(bbStringFromCString("Unicode character out of UCS-2 range"), BBNULL);
#					endif
						bbExThrow(ex);
					}
					
					*str_chr++ = v;
				}
			}
		}
	}
	
	BBString *str = bbStringFromShorts(str_buf, str_chr - str_buf);
	free(str_buf);
	
	return str;
}

BBString* bbStringToString(const BBString *str)
{
	return str;
}

BBInt bbStringCompare(const BBString *str, const BBString *rhs)
{
	if (str == rhs) {
		return 0;
	}
	
	const BBSize len = BB_MIN(str->length, rhs->length);
	
	BBInt difference;
	for (strpos_t i = 0; i < len; ++i)
	{
		if (difference = str->buf[i] - rhs->buf[i]) {
			return difference;
		}
	}
	
	return str->length - rhs->length;
}

BBBool bbStringStartsWith(const BBString *str, const BBString *sub)
{
	const BBSize sub_len = sub->length;
	
	if (str->length < sub_len) {
		return BBFALSE;
	}
	
	const BBChar *str_buf = str->buf - 1;
	const BBChar *sub_buf = sub->buf - 1;
	
	for (strpos_t i = 0; i < sub_len; ++i)
	{
		if (*++str_buf != *++sub_buf) {
			return BBFALSE;
		}
	}
	
	return BBTRUE;
}

BBBool bbStringEndsWith(const BBString *str, const BBString *sub)
{
	const BBSize len = sub->length;
	
	if (str->length < len) {
		return BBFALSE;
	}
	
	const BBChar *str_buf = str->buf + str->length - len - 1;
	const BBChar *sub_buf = sub->buf - 1;
	
	for (strpos_t i = 0; i < len; ++i)
	{
		if (*++str_buf != *++sub_buf) {
			return BBFALSE;
		}
	}
	
	return BBTRUE;
}

BBBool bbStringContains(const BBString *str, const BBString *sub)
{
	const BBSize sub_len = sub->length;
	
	if (sub_len <= 0 || sub_len > str->length) {
		return BBFALSE;
	}
	
	const BBChar *str_buf = str->buf - 1;
	const BBChar *sub_buf = sub->buf;
	
	const BBSize cmp_len = str->length - sub_len + 1;
	
	strpos_t cmp_pos;
	for (strpos_t i = 0; i < cmp_len; ++i)
	{
		for (cmp_pos = 0; cmp_pos < sub_len; ++cmp_pos)
		{
			if (*++str_buf != sub_buf[cmp_pos]) {
				break;
			}
		}
		
		if (cmp_pos == sub_len) {
			return BBTRUE;
		}
	}
	
	return BBFALSE;
}

BBString* bbStringConcat(const BBString *str, const BBString *rhs)
{
	uintmax_t out_len = str->length + rhs->length;
	
	if (out_len < 1) {
		return &bbEmptyString;
	} else if (out_len > BBSTRING_MAX_LENGTH) {
		out_len = BBSTRING_MAX_LENGTH;
	}
	
    BBString *out = BBSTRING_ALLOC(out_len);
	out->length = out_len;
	
    memcpy(out->buf, str->buf, str->length * sizeof (BBChar));
    memcpy(out->buf + str->length, rhs->buf, rhs->length * sizeof (BBChar));
	
    return out;
}

BBString* bbStringSlice(const BBString *str, BBInt begin, BBInt end)
{
	if (end - begin < 1) {
		return &bbEmptyString;
	}
	
	BBString *out = bbStringNew(end - begin);
	
	BBChar *out_buf = out->buf;
	BBSize str_len = str->length;
	
	int k, n;
	if ((n = -begin) > 0)
	{
		if (begin + n > end) {
			n = end - begin;
		}
		
		for (k = 0; k < n; ++k) {
			*out_buf++ = ' ';
		}
		
		if ((begin += n) == end) {
			return out;
		}
	}
	
	if ((n = str_len - begin) > 0)
	{
		BBChar *q = str->buf + begin;
		
		if (begin + n > end) {
			n = end - begin;
		}
		
		for (k = 0; k < n; ++k) {
			*out_buf++ = *q++;
		}
		
		if ((begin += n) == end) {
			return out;
		}
	}
	
	if ((n = end - begin) > 0)
	{
		for (k = 0; k < n; ++k)  {
			*out_buf++ = ' ';
		}
	}
	
	return out;
}

BBString* bbStringTrim(const BBString *str, BBStringSide side)
{
	const BBSize str_len = str->length;
	
	if (!str_len) {
		return &bbEmptyString;
	}
	
	if (side < BBSTRING_SIDE_LEFT || side > BBSTRING_SIDE_RIGHT) {
		side = BBSTRING_SIDE_BOTH;
	}
	
	const BBChar *str_buf = str->buf - 1;
	
	BBSize out_len  = str_len;
	
	strpos_t left_pos;
	if (side <= BBSTRING_SIDE_BOTH) {
		
		for (left_pos = 0; left_pos < str_len; ++left_pos)
		{
			if (!isspace(*++str_buf)) {
				break;
			}
		}
		
		out_len = str_len - left_pos;
	}
	
	if (side >= BBSTRING_SIDE_BOTH) {
		
		str_buf = str->buf + str_len;
		
		for (; out_len > 0; --out_len)
		{
			if (!isspace(*--str_buf)) {
				break;
			}
		}
	}
	
	BBString *out	= BBSTRING_NEW(out_len);	
	BBChar *out_buf	= out->buf - 1;
	
	str_buf = str->buf + left_pos - 1;
	
	for (strpos_t i = 0; i < out_len; ++i)
	{
		*++out_buf = *++str_buf;
	}
	
	return out;
}

BBString* bbStringPad(const BBString *str, BBInt len, BBStringSide side)
{
	const BBSize str_len = str->length;
	
	if (len <= str_len) {
		return str;
	} else if (len > BBSTRING_MAX_LENGTH) {
		len = BBSTRING_MAX_LENGTH;
	}
	
	if (side < BBSTRING_SIDE_LEFT || side > BBSTRING_SIDE_RIGHT) {
		side = BBSTRING_SIDE_BOTH;
	}
	
	BBString *out	= BBSTRING_NEW(len);
	BBChar *out_buf	= out->buf;
	
	// Identifies the length to pad now.
	len -= str_len;
	
	strpos_t pos, pad_end;
	
	if (side <= BBSTRING_SIDE_BOTH) {
		
		if (side == BBSTRING_SIDE_LEFT) {
			pad_end = len;
		} else {
			pad_end = ceilf(len / 2.0f);
		}
		
		--out_buf;
		
		for (pos = 0; pos <= pad_end; ++pos)
		{
			*++out_buf = ' ';
		}
	}
	
	--out_buf;	
	
	for (pos = 0; pos <= str_len; ++pos)
	{
		*++out_buf = str->buf[pos];
	}
	
	if (side >= BBSTRING_SIDE_BOTH) {
		
		--out_buf;
		for (pos = 0; pos < len; ++pos)
		{
			*++out_buf = ' ';
		}
	}
	
	return out;
}

BBString* bbStringReplace(const BBString *str, const BBString *sub, const BBString *replacement)
{
	if (!sub->length) {
		return str;
	}
	
	BBSize num_matches = bbStringCount(str, sub);
	
	if (!num_matches) {
		return str;
	}
	
	// todo: check for too large negatives
	int len_diff	= replacement->length - sub->length;
	int out_len		= str->length + (len_diff * num_matches);
	
	BB_ASSERT(out_len < BBSTRING_MAX_LENGTH, "The resulting string would be too long.");
	
	if (out_len < 1) {
		return &bbEmptyString;
	} else if (out_len > BBSTRING_MAX_LENGTH) {
		out_len = BBSTRING_MAX_LENGTH;
	}
	
	BBString *out	= BBSTRING_NEW(out_len);
	BBChar *out_buf	= out->buf;
	
	BBSize last_pos	= 0;	
	BBSize pos		= bbStringFind(str, sub, 0);
	
	while (pos != BBSTRING_NOTFOUND)
	{	
		len_diff = pos - last_pos;
	
		if (len_diff) {
			bb_strbuf_copy(str->buf + last_pos, out_buf, len_diff);
			out_buf += len_diff;
		}
		
		bb_strbuf_copy(replacement->buf, out_buf, replacement->length);
		out_buf += replacement->length;
		
		last_pos	= pos + sub->length;
		pos			= bbStringFind(str, sub, last_pos);
	}
	
	len_diff = str->length - last_pos;
	
	if (len_diff) {
		bb_strbuf_copy(str->buf + last_pos, out_buf, len_diff);
	}
	
	return out;
}

BBInt bbStringAsc(const BBString *str)
{
	return str->length ? str->buf[0] : BBSTRING_NOTFOUND;
}

BBInt bbStringFind(const BBString *str, const BBString *sub, BBInt index)
{
	if (str == sub) {
		return 0;
	}
	
	if (index < 0) {
		index = 0;
	}
	
	if (index + sub->length > str->length) {
		return BBSTRING_NOTFOUND;
	}
	
	const BBSize end_index = str->length - sub->length;
	
	while (index <= end_index)
	{
		if (bb_strbuf_equals(str->buf + index, sub->buf, sub->length)) {
			return index;
		}
		++index;
	}

	return BBSTRING_NOTFOUND;
}

BBInt bbStringFindLast(const BBString *str, const BBString *sub, BBInt index)
{
	BB_ASSERT(index >= 0, "Index must be positive or 0.");

	if (str == sub) {
		return 0;
	}
	
	index = str->length - index;
	
	if (index + sub->length > str->length) {
		index = str->length - sub->length;
	}
	
	while (index >= 0)
	{
		if (bb_strbuf_equals(str->buf + index, sub->buf, sub->length)) {
			return index;
		}
		
		--index;
	}
	
	return BBSTRING_NOTFOUND;
}


//	EXTENDED //

BBArray* bbStringFindAll(const BBString *str, const BBString *sub, BBInt index)
{
	if (index < 0) {
		index = 0;
	}
	
	if (sub->length + index > str->length) {
		return &bbEmptyArray;
	}

	// Same memory location means it's the same string, therefore one match.	
	if (str == sub) {
		BBIntArray *out	= bb_array_alloc1d("i", 1);
		out->data[0]	= 0;
		return out;
	}
	
	BBSize matches[(str->length / sub->length) + 1];
	BBSize num_matches = 0;
	
	const strpos_t last_search_pos = str->length - sub->length + 1;
	
	for (strpos_t i = index; i < last_search_pos; ++i)
	{
		if (bb_strbuf_equals(str->buf + i, sub->buf, sub->length)) {
			matches[num_matches++] = i;
		}
	}
	
	if (!num_matches) {
		return &bbEmptyArray;
	}
	
	BBIntArray *out = bb_array_alloc1d("i", num_matches);
	memcpy(out->data, matches, num_matches * sizeof (BBInt));
	
	return out;
}

BBInt bbStringFindAny(const BBString *str, const BBStringArray *arr, BBInt index)
{
	const BBSize str_len = str->length;
	const BBSize arr_len = arr->scales[0];	
	
	if (!str_len || !arr_len || arr->type[0] != '$') {
		return BBSTRING_NOTFOUND;
	}
	
	if (index < 0) {
		index = 0;
	} else if (index > str_len) {
		return BBSTRING_NOTFOUND;
	}
	
	for (strpos_t str_pos = index; str_pos < str_len; ++str_pos)
	{
		for (arrpos_t arr_pos = 0; arr_pos < arr_len; ++arr_pos)
		{
			BBString *sub = arr->data[arr_pos];
			
			if (sub->length < 1 || sub->length > (str_len - str_pos)) {
				continue;
			}
			
			if (str == sub || bb_strbuf_equals(str->buf + str_pos, sub->buf, sub->length)) {
				return str_pos;
			}
		}
	}
	
	return BBSTRING_NOTFOUND;
}


BBInt bbStringToInt(const BBString *str)
{
	const BBSize str_len = str->length;
	
	int i;
	for (i = 0; i < str_len; ++i)
	{
		if (!isspace(str->buf[i])) {
			break;
		}
	}
	
	bool is_negative = 0;	
	int num = 0;
	
	if (i == str_len) {
		return 0;
	}
	if (str->buf[i] == '+') {
		++i;
	} else if (is_negative = (str->buf[i] == '-')) {
		++i;
	}
	
	if (i == str_len) {
		return 0;
	}
	
	if (str->buf[i] == '%') {
		
		for (++i; i < str_len; ++i)
		{
			int c = str->buf[i];
			if (c != '0' && c != '1') {
				break;
			}
			
			num = num * 2 + (c - '0');
		}
		
	} else if (str->buf[i] == '$') {
		
		for (++i; i < str_len; ++i)
		{
			int c = toupper(str->buf[i]);
			
			if (!isxdigit(c)) {
				break;
			}
			
			if (c >= 'A') {
				c -= ('A' - '0' - 10);
			}
			
			num = num * 16 + (c - '0');
		}
		
	} else {
		
		for (; i < str_len; ++i)
		{
			int c = str->buf[i];
			if (!isdigit(c)) {
				break;
			}
			
			num = num * 10 + (c - '0');
		}
	}
	
	return is_negative ? -num : num;
}

void bbStringToLong(const BBString *str, BBLong *result)
{
	const BBSize str_len = str->length;
	
	int i;
	for (i = 0; i < str_len; ++i)
	{
		if (!isspace(str->buf[i])) {
			break;
		}
	}
	
	// String only contains spaces.
	if (i == str_len) {
		*result = 0;
		return;
	}

	bool is_negative = false;	
	
	if (str->buf[i] == '+') {
		++i;
	} else if (is_negative = (str->buf[i] == '-')) {
		++i;
	}
	
	// String only contains a sign operator.
	if (i == str_len) {
		*result = 0;
		return;
	}
	
	BBLong num = 0;	
	
	if (str->buf[i] == '%') {
		// Parse binary number.
		for (++i; i < str_len; ++i)
		{
			int c = str->buf[i];
			if (c != '0' && c != '1') {
				break;
			}
			num = num * 2 + (c - '0');
		}
	} else if (str->buf[i] == '$') {
		// Parse hexadecimal number.
		for (++i; i < str_len; ++i)
		{
			int c = toupper(str->buf[i]);
			if (!isxdigit(c)) {
				break;
			}
			if (c >= 'A') {
				c -= ('A' - '0' - 10);
			}
			
			num = num * 16 + (c - '0');
		}
	} else {
		// Parse decimal number.
		for (; i < str_len; ++i)
		{
			int c = str->buf[i];
			if (!isdigit(c)) {
				break;
			}
			num = num * 10 + (c - '0');
		}
	}
	
	*result = is_negative ? -num : num;
}

BBFloat bbStringToFloat(const BBString *str)
{
	char *c_str = bbStringToCString(str);
	float num = atof(c_str);
	bbMemFree(c_str);
	return num;
}

BBDouble bbStringToDouble(const BBString *str)
{
	char *p = bbStringToCString(str);
	double num = atof(p);
	bbMemFree(p);
	return num;
}

BBString* bbStringToLower(const BBString *str)
{
	const BBSize len = str->length;
	
	if (!len) {
		return &bbEmptyString;
	}
	
	BBString *out = BBSTRING_NEW(len);
	
	for (strpos_t k = 0; k < len; ++k)
	{
		int chr = str->buf[k];
		
		if (chr < 192) {
			chr = (chr >= 'A' && chr <= 'Z') ? (chr | 32) : chr;
		} else {
			
			int low		= 0;
			int high	= sizeof (bbToLowerData) / 4 - 1;
			
			while (low <= high)
			{
				int mid = (low + high) / 2;
				
				if (chr < bbToLowerData[mid * 2]) {
					high = mid - 1;
				} else if (chr > bbToLowerData[mid * 2]) {
					low  = mid + 1;
				} else {
					chr = bbToLowerData[mid * 2 + 1];
					break;
				}
			}
		}
		
		out->buf[k] = chr;
	}
	
	return out;
}

BBString* bbStringToUpper(const BBString *str)
{
	const BBSize len = str->length;
	
	if (!len) {
		return &bbEmptyString;
	}
	
	BBString *out = BBSTRING_NEW(len);
	
	for (strpos_t k = 0; k < str->length; ++k)
	{
		int chr = str->buf[k];
		
		if (chr < 181) {
			chr = (chr >= 'a' && chr <= 'z') ? (chr & ~32) : chr;
		} else {
			
			int low		= 0;
			int high	= sizeof (bbToUpperData) / 4 - 1;
			
			while (low <= high)
			{
				int mid = (low + high) / 2;
				
				if (chr < bbToUpperData[mid * 2]) {
					high = mid - 1;
				} else if (chr > bbToUpperData[mid * 2]) {
					low = mid + 1;
				} else {
					chr = bbToUpperData[mid * 2 + 1];
					break;
				}
			}
		}
		
		out->buf[k] = chr;
	}
	
	return out;
}

char* bbStringToCString(const BBString *str)
{
	const BBSize len = str->length;
	char *buf = (char *)bbMemAlloc(len + 1);
	
	for (strpos_t k = 0; k < len; ++k)
	{
		buf[k] = str->buf[k];
	}
	
	buf[len] = 0;
	
	return buf;
}

BBChar* bbStringToWString(const BBString *str)
{
	const BBSize len = str->length;
	BBChar *buf = (BBChar *)bbMemAlloc((len + 1) * sizeof (BBChar));
	
	memcpy(buf, str->buf, len * sizeof (BBChar));
	buf[len] = 0;
	
	return buf;
}

char* bbStringToUTF8String(const BBString *str)
{
	const BBSize str_len = str->length;
	
	char *out		= (char *)bbMemAlloc(str_len * 3 + 1);
	char *out_buf	= out;
	
	const BBChar *str_buf = str->buf - 1;
	
	for (strpos_t i = 0; i < str_len; ++i)
	{
		uint32_t code_point = *++str_buf;
		
		if (code_point < 0x80) {
			*out_buf++ = code_point;
		} else if (code_point < 0x800) {
			*out_buf++ = 0xc0 | (code_point >> 6);
			*out_buf++ = 0x80 | (code_point & 0x3f);
		} else {
			*out_buf++ = 0xe0 | (code_point >> 12);
			*out_buf++ = 0x80 | ((code_point >> 6) & 0x3f);
			*out_buf++ = 0x80 | (code_point & 0x3f);
		}
	}
	
	*out_buf = 0;
	return out;
}

BBStringArray* bbStringSplit(const BBString *str, const BBString *sep)
{
	int i, i2;
	
	BBSize num_bits = 0;
	
	BBString		*bit;
	BBStringArray	*bits;
	BBString		**bits_data;
	
	if (sep->length) {
		
		i = 0; num_bits = 1;
		
		while ((i2 = bbStringFind(str, sep, i)) != BBSTRING_NOTFOUND)
		{
			++num_bits;
			i = i2 + sep->length;
		}
		
		bits		= bb_array_alloc1d("$", num_bits);
		bits_data	= bits->data;
	
		i = 0;
		while (num_bits--)
		{
			i2 = bbStringFind(str, sep, i);
			if (i2 == BBSTRING_NOTFOUND) {
				i2 = str->length;
			}
			
			bit = bbStringSlice(str, i, i2);
			BB_INCREFS(bit);
			*bits_data++ = bit;
			i = i2 + sep->length;
		}
		
		return bits;
	}
	
	i = 0;
	
	for (;;)
	{
		while (i != str->length && str->buf[i] < 33)
		{
			++i;
		}
		
		if (i++ == str->length) {
			break;
		}
		
		while (i != str->length && str->buf[i] > 32)
		{
			++i;
		}
		
		++num_bits;
	}
	
	if (!num_bits) {
		return &bbEmptyArray;
	}
		
	bits		= bb_array_alloc1d("$", num_bits);
	bits_data	= bits->data;
	
	i = 0;
	while (num_bits--)
	{
		while (str->buf[i] < 33) {
			++i;
		}
		
		i2 = i++;
		
		while (i != str->length && str->buf[i] > 32) {
			++i;
		}
		
		bit = bbStringSlice(str, i2, i);
		BB_INCREFS(bit);
		*bits_data++ = bit;
	}
	
	return bits;
}

/*
BBStringArray* bbStringSplit(const BBString *str, const BBString *sep)
{
	int i, i2;
	BBString *bit;

	BBStringArray *out;
	BBString **out_data;
	
	if (sep->length) {
		
		BBSize matches[BBSTRING_MAX_MATCHES(str, sep)];
		BBSize num_matches;
		
		num_matches = bb_strbuf_findall(str->buf, str->length, sep->buf, sep->length, matches);
		
		// No matches found, return input string as 1d array
		//if (bit_matches->scales[0] < 1) {
		if (!num_matches) {
			out = bb_array_alloc1d("$", 1);
			out->data[0] = str;
			BB_RETAIN(out->data[0]);	// kinda dirty, but bbStringClone is overkill
			return out;
		}
		
		++num_matches;
		
		out = bb_array_alloc1d("$", num_matches);
		out_data = out->data;

		i = 0;
		while (num_matches--)
		{
			i2 = bbStringFind(str, sep, i);
			if (i2 == BBSTRING_NOTFOUND) {
				i2 = str->length;
			}
			
			bit = bbStringSlice(str, i, i2);
			BB_RETAIN(bit);
			*out_data++ = bit;
			
			i = i2 + sep->length;
		}

	} else {	
		i = 0; BBInt num_bits = 0;
		
		for (;;)
		{
			while (i != str->length && str->buf[i] < 33) {
				++i;
			}
			
			if (i++ == str->length) {
				break;
			}
			
			while (i != str->length && str->buf[i] > 32) {
				++i;
			}
			
			++num_bits;
		}
		
		if (!num_bits) {
			return &bbEmptyArray;
		}
		
		out = bb_array_alloc1d("$", num_bits);
		out_data = out->data;
		i = 0;
		
		while (num_bits--)
		{
			while (str->buf[i] < 33)
			{
				++i;
			}
			
			i2 = i++;
			
			while (i != str->length && str->buf[i] > 32)
			{
				++i;
			}
			
			bit = bbStringSlice(str, i2, i);
			BB_RETAIN(bit);
			*out_data++ = bit;
		}
	}
		
	return out;
}
*/

BBString* bbStringJoin(const BBString *sep, const BBStringArray *bits)
{
	if (bits == &bbEmptyArray) {
		return &bbEmptyString;
	}

	const BBSize num_bits	= bits->scales[0];
	BBSize bit_id			= 0;
	BBSize size				= 0;
	
	BBString **bits_data = (BBString **)BBARRAY_DATAPTR_CALC(bits, 1);
	
	for (bit_id = 0; bit_id < num_bits; ++bit_id)
	{
		size += (*bits_data++)->length;
		//BBString *bit	= *bits_data++;
		//size			+= bit->length;
	}
	
	size += (num_bits - 1) * sep->length;
	
	// Checking for "size < 1" is not necessary.
	BBString *str	= BBSTRING_NEW(size);
	BBChar *str_buf	= str->buf;
	
	bits_data = (BBString **)BBARRAY_DATAPTR_CALC(bits, 1);
	
	for (bit_id = 0; bit_id < num_bits; ++bit_id)
	{
		if (bit_id) {
			memcpy(str_buf, sep->buf, sep->length * sizeof (BBChar));
			str_buf += sep->length;
		}
		
		BBString *bit = *bits_data++;
		
		memcpy(str_buf, bit->buf, bit->length * sizeof (BBChar));
		str_buf += bit->length;
	}
	
	return str;
}

static void mktmp(void *ptr)
{
	static int i;
	static void *bufs[32];
	int n = bbAtomicAdd(&i, 1) & 31;
	bbMemFree(bufs[n]);
	bufs[n] = ptr;
}

char* bbTmpCString(const BBString *str)
{
	char *out = bbStringToCString(str);
	mktmp(out);
	return out;
}

BBChar* bbTmpWString(const BBString *str)
{
	BBChar *out = bbStringToWString(str);
	mktmp(out);
	return out;
}

char* bbTmpUTF8String(const BBString *str)
{
	char *out = bbStringToUTF8String(str);
	mktmp(out);
	return out;
}


//	EXTENDED //

BBString* bbStringClone(const BBString *str)
{
	const BBSize str_len = str->length;
	
	if (!str_len) {
		return &bbEmptyString;
	}
	
	BBString *out = BBSTRING_NEW(str_len);
	
	const BBChar *str_buf;
	BBChar *out_buf;
	
	str_buf = str->buf - 1;
	out_buf = out->buf - 1;
	
	for (strpos_t i = 0; i < str_len; ++i)
	{
		*++out_buf = *++str_buf;
	}
	
	return out;
}

BBBool bbStringEquals(const BBString *str, const BBString *rhs)
{
	if (str == rhs) {
		return BBTRUE;
	}
	
	if (str->length != rhs->length) {
		return BBFALSE;
	}

	const BBSize len			= str->length;	
	const BBChar *str_buf	= str->buf - 1;
	const BBChar *rhs_buf	= rhs->buf - 1;
	
	for (strpos_t i = 0; i < len; ++i)
	{
		if (*++str_buf != *++rhs_buf) {
			return BBFALSE;
		}
	}
	
	return BBTRUE;
}

BBInt bbStringCount(const BBString *str, const BBString *sub)
{
	if (str == sub) {
		return 1;
	}
	
	const BBSize sub_len = sub->length;
	const BBSize search_len = str->length - sub_len + 1;
	
	if (search_len < 1 || sub_len < 1) {
		return 0;
	}
	
	BBInt count = 0;
	
	for (strpos_t i = 0; i < search_len; ++i)
	{
		if (bb_strbuf_equals(str->buf + i, sub->buf, sub_len)) {
			++count;
		}
	}
	
	return count;
}

BBString* bbStringReverse(const BBString *str)
{
	if (!str->length) {
		return &bbEmptyString;
	}	
	
	const BBSize len = str->length;
	
	BBString *out = BBSTRING_ALLOC(len);
	out->length = len;
	
	const BBChar *str_buf = str->buf + len;
	BBChar *out_buf = out->buf - 1;
	
	for (strpos_t i = 0; i < len; ++i)
	{
		*++out_buf = *--str_buf;
	}
	
	return out;
}

BBString* bbStringRotate(const BBString *str, BBInt count)
{
	const BBSize str_len = str->length;
	
	if (str_len < 2 || !count || count == str_len) {
		return str;
	}
	
	count = -(count % str_len);
	
	if (count < 0) {
		count += str_len;
	}
		
	BBString *out	= BBSTRING_ALLOC(str_len);
	out->length		= str_len;
	
	for (strpos_t i = 0; i < str_len; ++i, ++count)
	{
		if (count >= str_len) {
			count = 0;
		}
		
		out->buf[i] = str->buf[count];
	}
		
	return out;
}

BBString* bbStringTimes(const BBString *str, BBInt count)
{
	const BBSize str_len	= str->length;
	
	// Set to maximum count value for testing (temporarily)
	BBSize out_len = BBSTRING_MAX_LENGTH / str_len;
		
	if (count < 2) {
		return count == 1 ? str : &bbEmptyString;
	} else if (count > out_len) {
		count = out_len;
	}
	
	// Set to output string's length
	out_len = str_len * count;
		
	BBString *out	= BBSTRING_NEW(out_len);
	BBChar *out_buf	= out->buf - 1;
	
	const BBChar *str_buf = str->buf;
	
	for (strpos_t out_pos = 0; out_pos < count; ++out_pos)
	{
		for (strpos_t str_pos = 0; str_pos < str_len; ++str_pos) {
			*++out_buf = str_buf[str_pos];
		}
	}
	
	return out;
}

BBString* bbStringSub(const BBString *str, BBInt index, BBInt len)
{
	if (!str->length || str->length < index || len < 1) {
		return &bbEmptyString;
	}
	
	if (index < 0) {
		index += str->length;
		
		if (index + len <= 0) {
			return &bbEmptyString;
		}
	}
	
	if (index + len > str->length) {
		len = str->length - index;
	}
	
	BBString *out = BBSTRING_NEW(len);
	
	const BBChar *str_buf = str->buf + index - 1;
	BBChar *out_buf = out->buf - 1;
	
	for (strpos_t i = 0; i < len; ++i)
	{
		*++out_buf = *++str_buf;
	}
	
	return out;
}

BBBool bbStringIsLower(const BBString *str)
{
	const BBSize len = str->length;
	
	if (!len) {
		return BBTRUE;
	}
	
	for (strpos_t pos = 0; pos < len; ++pos)
	{
		int chr = str->buf[pos];
		
		if (chr < 192) {
			chr = (chr >= 'A' && chr <= 'Z') ? (chr | 32) : chr;
		} else {
			int low  = 0;
			int high = sizeof (bbToLowerData)/4 - 1;
			
			while (low <= high)
			{
				int mid = (low + high) / 2;
				
				if (chr < bbToLowerData[mid * 2]) {
					high = mid - 1;
				} else if (chr > bbToLowerData[mid * 2]) {
					low  = mid + 1;
				} else {
					chr = bbToLowerData[mid * 2 + 1];
					break;
				}
			}
		}
		
		if (chr != str->buf[pos]) {
			return BBFALSE;
		}
	}
	
	return BBTRUE;
}

BBBool bbStringIsUpper(const BBString *str)
{
	const BBSize len = str->length;
	
	if (!len) {
		return BBTRUE;
	}
	
	for (strpos_t pos = 0; pos < str->length; ++pos)
	{
		int chr = str->buf[pos];
		
		if (chr < 181) {
			chr = (chr >= 'a' && chr <= 'z') ? (chr & ~32) : chr;
		} else {
			int low = 0;
			int high = sizeof (bbToUpperData) / 4 - 1;
			
			while (low <= high)
			{
				int mid = (low + high) / 2;
				
				if (chr < bbToUpperData[mid * 2]) {
					high = mid - 1;
				} else if (chr > bbToUpperData[mid * 2]) {
					low = mid + 1;
				} else {
					chr = bbToUpperData[mid * 2 + 1];
					break;
				}
			}
		}
		
		if (chr != str->buf[pos]) {
			return BBFALSE;
		}
	}
	
	return BBTRUE;
}

BBBool bbStringIsASCII(const BBString *str)
{
	const BBSize len = str->length;
	
	if (!len) {
		return BBTRUE;
	}
	
	const BBChar * const str_buf = str->buf;
	
	for (strpos_t pos = 0; pos < len; ++pos)
	{
		if (str_buf[pos] > 127) {
			return BBFALSE;
		}
	}
	
	return BBTRUE;
}
